@@ -30,17 +30,23 @@ namespace GtkSharp.Generation {
 		ArrayList parms = new ArrayList ();
 		ArrayList special = new ArrayList ();
 		string error_param = null;
+		string user_data_param = null;
+		string destroy_param = null;
 
-		public ManagedCallString (Parameters parms)
+		public ManagedCallString (Parameters parms, bool drop_first)
 		{
-			for (int i = 1; i < parms.Count; i ++) {
+			for (int i = drop_first ? 1 : 0; i < parms.Count; i ++) {
 				Parameter p = parms [i];
-				if (p.IsLength && parms [i-1].IsString) 
+				if (p.IsLength && i > 0 && parms [i-1].IsString) 
 					continue;
-				else if (p.Scope == "notified")
+				else if (p.Scope == "notified") {
+					user_data_param = parms[i+1].Name;
+					destroy_param = parms[i+2].Name;
 					i += 2;
-					
-				else if (p is ErrorParameter) {
+				} else if (p.IsUserData && (i == parms.Count - 1) && (parms.HideData || parms [i-1].Generatable is CallbackGen)) {
+					user_data_param = p.Name;
+					continue;
+				} else if (p is ErrorParameter) {
 					error_param = p.Name;
 					continue;
 				}
@@ -55,13 +61,27 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		public bool HasOutParam {
+			get {
+				foreach (Parameter p in parms) {
+					if (p.PassAs == "out")
+						return true;
+				}
+				return false;
+			}
+		}
+
+		public string Unconditional (string indent) {
+			string ret = "";
+			if (error_param != null)
+				ret = indent + error_param + " = IntPtr.Zero;\n";
+			return ret;
+		}
+
 		public string Setup (string indent)
 		{
 			string ret = "";
 
-			if (error_param != null)
-				ret = indent + error_param + " = IntPtr.Zero;\n";
-
 			for (int i = 0; i < parms.Count; i ++) {
 				if ((bool)special[i] == false)
 					continue;
@@ -70,7 +90,12 @@ namespace GtkSharp.Generation {
 				IGeneratable igen = p.Generatable;
 
 				if (igen is CallbackGen) {
-					ret += indent + String.Format ("{0} {1}_invoker = new {0} ({1});\n", (igen as CallbackGen).InvokerName, p.Name);
+					if (user_data_param == null)
+						ret += indent + String.Format ("{0} {1}_invoker = new {0} ({1});\n", (igen as CallbackGen).InvokerName, p.Name);
+					else if (destroy_param == null)
+						ret += indent + String.Format ("{0} {1}_invoker = new {0} ({1}, {2});\n", (igen as CallbackGen).InvokerName, p.Name, user_data_param);
+					else
+						ret += indent + String.Format ("{0} {1}_invoker = new {0} ({1}, {2}, {3});\n", (igen as CallbackGen).InvokerName, p.Name, user_data_param, destroy_param);
 				} else {
 					ret += indent + igen.QualifiedName + " my" + p.Name;
 					if (p.PassAs == "ref")
