@@ -24,6 +24,7 @@ namespace GLib {
 
 	using System;
 	using System.Collections;
+	using System.Reflection;
 	using System.Runtime.InteropServices;
 
 	[StructLayout (LayoutKind.Sequential)]
@@ -57,6 +58,16 @@ namespace GLib {
 			g_value_set_boolean (ref this, val);
 		}
 
+		public Value (byte val) : this (GType.UChar)
+		{
+			g_value_set_uchar (ref this, val);
+		}
+
+		public Value (sbyte val) : this (GType.Char)
+		{
+			g_value_set_char (ref this, val);
+		}
+
 		public Value (int val) : this (GType.Int)
 		{
 			g_value_set_int (ref this, val);
@@ -133,6 +144,10 @@ namespace GLib {
 			g_value_set_object (ref this, val == null ? IntPtr.Zero : val.Handle);
 		}
 
+		public Value (GLib.GInterfaceAdapter val) : this (val == null ? GType.Object : val.GType)
+		{
+			g_value_set_object (ref this, val == null ? IntPtr.Zero : val.Handle);
+		}
 
 		public Value (GLib.Object obj, string prop_name)
 		{
@@ -204,6 +219,16 @@ namespace GLib {
 			return g_value_get_boolean (ref val);
 		}
 
+		public static explicit operator byte (Value val)
+		{
+			return g_value_get_uchar (ref val);
+		}
+
+		public static explicit operator sbyte (Value val)
+		{
+			return g_value_get_char (ref val);
+		}
+
 		public static explicit operator int (Value val)
 		{
 			return g_value_get_int (ref val);
@@ -309,7 +334,14 @@ namespace GLib {
 			Type t = GType.LookupType (type);
 			if (t == null)
 				throw new Exception ("Unknown type " + new GType (type).ToString ());
-			return Marshal.PtrToStructure (boxed_ptr, t);
+			else if (t.IsSubclassOf (typeof (GLib.Opaque)))
+				return (GLib.Opaque) this;
+
+			MethodInfo mi = t.GetMethod ("New", BindingFlags.Static | BindingFlags.Public | BindingFlags.FlattenHierarchy);
+			if (mi == null)
+				return Marshal.PtrToStructure (boxed_ptr, t);
+			else
+				return mi.Invoke (null, new object[] {boxed_ptr});
 		}
 
 		public object Val
@@ -317,6 +349,10 @@ namespace GLib {
 			get {
 				if (type == GType.Boolean.Val)
 					return (bool) this;
+				else if (type == GType.UChar.Val)
+					return (byte) this;
+				else if (type == GType.Char.Val)
+					return (sbyte) this;
 				else if (type == GType.Int.Val)
 					return (int) this;
 				else if (type == GType.UInt.Val)
@@ -336,18 +372,26 @@ namespace GLib {
 					return (string) this;
 				else if (type == GType.Pointer.Val)
 					return (IntPtr) this;
+				else if (type == GType.Param.Val)
+					return g_value_get_param (ref this);
 				else if (type == ManagedValue.GType.Val)
 					return ManagedValue.ObjectForWrapper (g_value_get_boxed (ref this));
 				else if (g_type_is_a (type, GType.Object.Val))
 					return (GLib.Object) this;
 				else if (g_type_is_a (type, GType.Boxed.Val))
 					return ToBoxed ();
+				else if (type == IntPtr.Zero)
+					return null;
 				else
 					throw new Exception ("Unknown type " + new GType (type).ToString ());
 			}
 			set {
 				if (type == GType.Boolean.Val)
 					g_value_set_boolean (ref this, (bool) value);
+				else if (type == GType.UChar.Val)
+					g_value_set_uchar (ref this, (byte) value);
+				else if (type == GType.Char.Val)
+					g_value_set_char (ref this, (sbyte) value);
 				else if (type == GType.Int.Val)
 					g_value_set_int (ref this, (int) value);
 				else if (type == GType.UInt.Val)
@@ -376,12 +420,17 @@ namespace GLib {
 					IntPtr buf = Marshal.AllocHGlobal (Marshal.SizeOf (value.GetType()));
 					Marshal.StructureToPtr (value, buf, false);
 					g_value_set_pointer (ref this, buf);
+				} else if (type == GType.Param.Val) {
+					g_value_set_param (ref this, (IntPtr) value);
 				} else if (type == ManagedValue.GType.Val) {
 					IntPtr wrapper = ManagedValue.WrapObject (value);
 					g_value_set_boxed (ref this, wrapper);
 					ManagedValue.ReleaseWrapper (wrapper);
 				} else if (g_type_is_a (type, GType.Object.Val))
-					g_value_set_object (ref this, ((GLib.Object) value).Handle);
+					if(value is GLib.Object)
+						g_value_set_object (ref this, (value as GLib.Object).Handle);
+					else
+						g_value_set_object (ref this, (value as GLib.GInterfaceAdapter).Handle);
 				else if (g_type_is_a (type, GType.Boxed.Val)) {
 					if (value is IWrapper) {
 						g_value_set_boxed (ref this, ((IWrapper)value).Handle);
@@ -395,6 +444,12 @@ namespace GLib {
 			}
 		}
 
+		internal void Update (object val)
+		{
+			if (g_type_is_a (type, GType.Boxed.Val) && !(val is IWrapper))
+				Marshal.StructureToPtr (val, g_value_get_boxed (ref this), false);
+		}
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern void g_value_init (ref GLib.Value val, IntPtr gtype);
 
@@ -413,6 +468,12 @@ namespace GLib {
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern void g_value_set_boolean (ref Value val, bool data);
 
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern void g_value_set_uchar (ref Value val, byte data);
+
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern void g_value_set_char (ref Value val, sbyte data);
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern void g_value_set_boxed (ref Value val, IntPtr data);
 
@@ -434,6 +495,9 @@ namespace GLib {
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern void g_value_set_object (ref Value val, IntPtr data);
 
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern void g_value_set_param (ref Value val, IntPtr data);
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern void g_value_set_pointer (ref Value val, IntPtr data);
 
@@ -451,6 +515,12 @@ namespace GLib {
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern bool g_value_get_boolean (ref Value val);
 
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern byte g_value_get_uchar (ref Value val);
+
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern sbyte g_value_get_char (ref Value val);
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern IntPtr g_value_get_boxed (ref Value val);
 
@@ -472,6 +542,9 @@ namespace GLib {
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern IntPtr g_value_get_object (ref Value val);
 
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern IntPtr g_value_get_param (ref Value val);
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern IntPtr g_value_get_pointer (ref Value val);
 
